import React from 'react';
import PropTypes from 'prop-types';
import { isEqual, noop } from 'lodash';

interface Props {
  id: string;
  onSizeChange: (key: string, size: [number, number]) => void;
  size: [number, number];
  validity: boolean;
  style?: React.CSSProperties;
}

export default class HorizontalItem extends React.Component<Props> {

  static defaultProps = {
    onWidthChange: noop,
    style: {},
    validity: false,
  };
  private mutationObserver: MutationObserver;

  constructor( props: Props ) {
    super( props );

    this.setRef = this.setRef.bind( this );
    this.updateSize = this.updateSize.bind( this );

    this.mutationObserver = new MutationObserver( this.updateSize );
    this.node = null;
  }

  componentDidMount() {
    if ( this.node ) {
      this.mutationObserver.observe( this.node, {
        childList: true,
        subtree: true,
        attributes: true,
        characterData: true,
      } );
      this.node.addEventListener( 'load', this.updateSize, true );
    }
    this.updateSize();
  }

  componentDidUpdate() {
    this.updateSize();
  }

  componentWillUnmount() {
    if ( this.mutationObserver ) {
      this.mutationObserver.disconnect();
      this.mutationObserver = null;
    }
  }

  shouldComponentUpdate( nextProps ) {
    if ( ! isEqual( nextProps.style, this.props.style ) ) return true;
    if (
      this.props.children &&
      nextProps.children &&
      ! isEqual( nextProps.children.props, this.props.children.props )
    )
      return true;
    if (
      this.props.children &&
      nextProps.children &&
      nextProps.children.key !== this.props.children.key
    )
      return true;
    if ( nextProps.width !== this.props.width ) return true;
    if ( ! nextProps.validity ) return true;
    if ( nextProps.hasFocus !== this.props.hasFocus ) return true;
    if ( nextProps.isMouseDownOnItem !== this.props.isMouseDownOnItem ) return true;
    if ( nextProps.hasFocusWithin !== this.props.hasFocusWithin ) return true;
    return false;
  }

  updateSize() {
    if ( ! this.node ) return;
    const r = this.node.getBoundingClientRect(),
      a = r.width, b = r.height;
    // console.log("called update width");
    ( Math.abs( a - this.props.size[0] ) > 0.5 || Math.abs( a - this.props.size[1] ) > 0.5 || ! this.props.validity ) &&
    this.props.onSizeChange( this.props.id, [a, b] );
  }

  setRef = r => {
    this.node = r;
  };

  render() {
    if ( ! this.props.children ) return null;
    return (
      <div
        key={ this.props.id }
        ref={ this.setRef }
        style={ Object.assign( this.props.style, {
          position: 'absolute',
        } ) }
      >
        { this.props.children }
      </div>
    );
  }
}
