import * as React from 'react';
import classNames from 'classnames';
import { noop, isEqual } from 'lodash';
// @ts-ignore
import {
  requestAnimationTimeout,
  cancelAnimationTimeout,
  // @ts-ignore
} from './requestAnimationTimeout';

import {Layout} from './layout';
import {ListItem} from './ListItem';
import {Scrollbar} from './ScrollBar';

// @ts-ignore
import gs from '@ladifire-css/sass/index.scss';

export interface VirtualizedListProps {
  width: number;
  height: number;
  className?: string;
  keys: any;
  layout: Layout;
  rowRenderer: (key: any) => JSX.Element;
  getAriaLabelForRow: (r: any) => string;
  role: string;
  itemRole: string;
  ariaLabel?: string;
  persistentKeys?: string[];
  fadeScrollbar: boolean;
  onScroll: (event: any, r?: any) => void;
  onScrollEnd: any;
  onKeyDown: any;
  onContentScroll: (event: any) => void;
  onSelectionChange: any;
  shouldHorizontallyScroll: boolean;
  useWindowScroller?: boolean;
  isScrolling?: boolean;
  scrollTop: number;
  onContentHeightChange?: (height: number) => void;
}

export interface VirtualizedListState {
  start: any;
  end: any;
  tops: any;
  scrolling: boolean;
}

export class List<T, K> extends React.PureComponent<any, any> {
  static defaultProps = {
    className: '',
    layout: void 0,
    role: 'list',
    itemRole: 'listitem',
    ariaLabel: undefined,
    persistentKeys: [],
    fadeScrollbar: false,
    onScroll: noop,
    onScrollEnd: noop,
    onContentScroll: noop,
    onSelectionChange: noop,
    getAriaLabelForRow: void 0,
    shouldHorizontallyScroll: false,
  };
  protected layout: Layout;
  protected node: any;
  private contentNode: any;
  private isSelecting: boolean;
  private selectionStart: any;
  private selectionEnd: any;
  public scrollStartTimeout: any;
  private _disablePointerEventsTimeoutId: any;
  public scrollTop: number | boolean;

  constructor( e: VirtualizedListProps ) {
    super( e );

    this.onScroll = this.onScroll.bind( this );
    this.onContentScroll = this.onContentScroll.bind( this );
    this.onSelectionChange = this.onSelectionChange.bind( this );
    this.setRef = this.setRef.bind( this );
    this.setContentRef = this.setContentRef.bind( this );
    this.setHeight = this.setHeight.bind( this );
    this.getScrollTop = this.getScrollTop.bind( this );
    this.getVisibleRange = this.getVisibleRange.bind( this );
    this.isItemInView = this.isItemInView.bind( this );

    this.layout = e ? e.layout : new Layout();
    this.node = null;
    this.contentNode = null;
    this.isSelecting = false;
    this.selectionStart = null;
    this.selectionEnd = null;
    this.scrollStartTimeout = null;
    this._disablePointerEventsTimeoutId = null;
    e && e.keys && this.layout.setKeys( e.keys );
    const n = this.getBounds( 0, e ? e.height : 0 ),
      s = n.start,
      l = n.end;
    const i = this.layout.getTops();
    this.state = {
      start: s,
      end: l,
      tops: i,
      scrolling: false,
    };
  }

  componentDidMount() {
    document.addEventListener( 'selectionchange', this.onSelectionChange );
  }

  UNSAFE_componentWillReceiveProps( r: VirtualizedListProps ) {
    r.width !== this.props.width && this.layout.heightCache.invalidate();
    isEqual( this.props.keys, r.keys ) || this.layout.setKeys( r.keys );
    this.relayout();
  }

  componentWillUnmount() {
    if ( this._disablePointerEventsTimeoutId ) {
      cancelAnimationTimeout( this._disablePointerEventsTimeoutId );
    }
    document.removeEventListener( 'selectionchange', this.onSelectionChange );
  }

  /**
   * Sets an :isScrolling flag for a small window of time.
   * This flag is used to disable pointer events on the scrollable portion of the Grid.
   * This prevents jerky/stuttery mouse-wheel scrolling.
   */
  _debounceScrollEnded() {
    const scrollingResetTimeInterval = 150; // {scrollingResetTimeInterval} = this.props || 50;

    if ( this._disablePointerEventsTimeoutId ) {
      cancelAnimationTimeout( this._disablePointerEventsTimeoutId );
    }

    this._disablePointerEventsTimeoutId = requestAnimationTimeout(
      this._debounceScrollEndedCallback,
      scrollingResetTimeInterval
    );
  }

  _debounceScrollEndedCallback = () => {
    if (this.props.onScrollEnd) {
      this.props.onScrollEnd();
    }
    this._disablePointerEventsTimeoutId = null;
    // isScrolling is used to determine if we reset styleCache
    this.setState( function( e: any ) {
      if ( ! e.scrolling ) return null;
      return {
        scrolling: false,
      };
    } );
  };

  onScroll( r: any, x?: any) {
    this._debounceScrollEnded();
    this.setState( function() {
      return {
        scrolling: true,
      };
    } );
    this.relayout();
    this.props.onScroll( r, {
      scrollTop: this.getScrollTop(),
      scrollHeight: this.getContentHeight(),
      clientHeight: this.props.height,
    } );
  }

  onContentScroll( r: any ) {
    this.props.onContentScroll( r );
  }

  onSelectionChange() {
    if ( ! this.node ) return;
    if (
      // @ts-ignore
      'Range' === window.getSelection().type &&
      this.contentNode &&
      // @ts-ignore
      this.contentNode.contains( window.getSelection().anchorNode )
    ) {
      this.isSelecting = true;
      this.selectionStart = this.props.keys[ this.state.start ];
      this.selectionEnd = this.props.keys[ this.state.end - 1 ];
    } else {
      this.isSelecting = false;
      this.selectionStart = null;
      this.selectionEnd = null;
    }
    this.props.onSelectionChange( this.isSelecting );
  }

  setRef( r: any ) {
    this.node = r;
  }

  setContentRef( r: any ) {
    this.contentNode = r;
  }

  getTop( r: any ) {
    return this.layout.getTop( r );
  }

  getBottom( r: any ) {
    return this.layout.getBottom( r );
  }

  getScrollTop() {
    if ( !this.props.useWindowScroller && ! this.node ) return null;
    return this.props.useWindowScroller ? this.props.scrollTop : this.node.scrollTop();
  }

  setHeight( r: string, t: number ) {
    if ( ! r ) return;
    if ( Math.abs( t - this.layout.getHeight( r ) ) < 0.5 ) return;
    this.layout.setHeight( r, t );
    this.relayout();
  }

  setScrollTop( r: number ) {
    if ( ! this.node ) return;
    this.node.scrollTop( Math.ceil( r ) );
  }

  getClassName() {
    return classNames(
      'c_virtual_list',
      gs.overflow_x__hidden,
      gs.overflow_y__hidden,
      this.props.className
    );
  }

  getContentsClassName() {
    return classNames( 'c_virtual_list__scroll_container', {
      [gs.pointer_event__none]: this.isScrolling() && ! this.isSelecting,
    } );
  }

  getVisibleRange() {
    const r =
      arguments.length > 0 && void 0 !== arguments[ 0 ] ? arguments[ 0 ] : this.getScrollTop();
    const t = this.props,
      n = t.keys,
      a = t.height;
    const s = this.layout.getBounds( r, r + a ),
      l = s.start,
      i = s.end;
    const o = this.layout.getTop( n[ l ] ) < r - 1 ? l + 1 : l;
    const u = this.layout.getBottom( n[ i - 1 ] ) > r + a + 1 ? i - 1 : i;
    const c = Math.min( o, n.length - 1 );
    const d = Math.max( u, c );
    return {
      start: c,
      end: d,
    };
  }

  getContentHeight() {
    return this.layout.getTotalHeight();
  }

  getContainerHeight() {
    return this.props.height;
  }

  getBounds( r: any, t: any ) {
    // this.onSelectionChange();
    let n = r;
    let a = t;
    this.selectionStart && ( n = Math.min( n, this.layout.getTop( this.selectionStart ) ) );
    this.selectionEnd && ( a = Math.max( a, this.layout.getBottom( this.selectionEnd ) ) );
    return this.layout.getBounds( n, a );
  }

  isScrolling() {
    return this.props.useWindowScroller ? this.props.isScrolling : this.state.scrolling;
  }

  isItemInView( r: string ) {
    const t = this.getScrollTop();
    const n = this.layout.getTop( r );
    return n >= t && n <= t + this.props.height;
  }

  isScrolledToTop() {
    return Math.floor( this.getScrollTop() ) <= 1;
  }

  isScrolledToBottom() {
    return this.getScrollTop() + this.getContainerHeight() >= this.layout.getTotalHeight() - 1;
  }

  scrollToKey( r: string, x?: any ) {
    const t = arguments.length > 1 && void 0 !== arguments[ 1 ] ? arguments[ 1 ] : {},
      n = t.lazy,
      a = t.animate;
    if ( n && this.isItemInView( r ) ) return;
    const s = this.layout.getOffsetForKey( r );
    this.scrollToOffset( s, {
      animate: a,
    } );
  }

  scrollToOffset( r: any, x?: any ) {
    const t = arguments.length > 1 && void 0 !== arguments[ 1 ] ? arguments[ 1 ] : {},
      n = t.animate;
    if ( ! this.node ) return;
    const a = this.bracketScrollTop( r );
    if ( n ) {
      this.node.scrollWithAnimation( a );
      return;
    }
    this.scrollTop = a;
    this.setScrollTop( a );
  }

  bracketScrollTop( r: number ) {
    return Math.max( 0, Math.min( this.getContentHeight() - this.props.height, r ) );
  }

  relayout() {
    var r = this;
    this.setState( function( e: any, t: any ) {
      var n = r.getScrollTop();
      var a = t.height / 2;
      var s = n - a;
      var l = n + t.height + a;
      var i = r.getBounds( s, l ),
        o = i.start,
        u = i.end;
      var c = r.layout.getTops();

      if ( o === e.start && u === e.end && c === e.tops ) return null;
      return {
        start: o,
        end: u,
        tops: c,
      };
    } );
  }

  renderItem( t: string, r: any ) {

    const n = this.getTop( r );
    const a = this.layout.getHeight( r );
    const s = this.layout.getHeightValidity( r );
    const l = this.props.getAriaLabelForRow;
    const i = l && l( t );

    return (
      <ListItem
        id={r}
        key={r}
        height={a}
        style={{
          top: n,
        }}
        role={this.props.itemRole}
        onHeightChange={this.setHeight}
        validity={s}
        ariaLabel={i}
        shouldHorizontallyScroll={this.props.shouldHorizontallyScroll}
      >
        {this.props.rowRenderer( t )}
      </ListItem>
    );
  }

  renderItems() {
    const r = this;
    const t = this.props,
      n = t.keys,
      a = t.persistentKeys;
    const s = this.state,
      l = s.start,
      i = s.end;
    const o = [];

    for ( let u = l; u < i; u++ ) {
      const c = n[ u ];
      c && o.push( this.renderItem( c, u ) );
    }
    a && a.forEach( function( e: any ) {
      const tt = n.indexOf( e );
      if ( tt < 0 ) return;
      if ( tt >= l && tt < i ) return;
      o.push( r.renderItem( e, tt ) );
    } );

    return o;
  }

  _handleWindowScroll = () => {
    // console.log('_handleWindowScroll:');
  };

  renderContents() {
    const r = this.getContentsClassName();
    const t = this.getContentHeight();
    const n = this.renderItems();

    return (
      <div
        ref={this.setContentRef}
        className={r}
        role='presentation'
        style={{
          position: 'relative',
          height: t,
        }}
      >
        {n}
      </div>
    );
  }

  render() {
    const r = this.props,
      t = r.width,
      n = r.height;
    if (!this.props.useWindowScroller && (0 === t || 0 === n) ) return null;
    const a = this.getClassName();
    const s = this.renderContents();

    if (this.props.useWindowScroller) {
      return (
        <div
          // onScroll={this._handleWindowScroll}
          // style={{height: this.getContentHeight()}}
          // style={{height: 15000}}
        >
          {s}
        </div>
      );
    }

    return (
      <Scrollbar
        className={a}
        width={t}
        height={this.getContainerHeight()}
        contentHeight={this.getContentHeight()}
        onTrackClick={Scrollbar.track.page}
        onScroll={this.onScroll}
        onContentScroll={this.onContentScroll}
        anchor='top'
        role={this.props.role}
        ref={this.setRef}
        ariaLabel={this.props.ariaLabel}
        fade={this.props.fadeScrollbar}
        onKeyDown={this.props.onKeyDown}
        isEmpty={!this.props.keys || this.props.keys.length === 0}
      >
        {s}
      </Scrollbar>
    );
  }
}
